from bloqade import start, var
from bloqade.serialize import dumps, loads
import numpy as np


def test_integration_1():
    (
        start.add_position((0, 0))
        .add_position((0, 5.0))
        .scale("r")
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, "ramp_time", 0.1], [-100, -100, 100, 100]
        )
        .amplitude.uniform.piecewise_linear([0.1, "ramp_time", 0.1], [0, 10, 10, 0])
        .assign(ramp_time=3.0)
        .batch_assign(r=np.linspace(4, 10, 11).tolist())
        .bloqade.python()
        .run(10000, cache_matrices=True, blockade_radius=6.0, interaction_picture=True)
        .report()
        .bitstrings()
    )


def test_integration_2():
    ramp_time = var("ramp_time")
    (
        start.add_position((0, 0))
        .add_position((0, 5.0))
        .scale("r")
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, "ramp_time", 0.1], [-100, -100, 100, 100]
        )
        .amplitude.uniform.piecewise_linear([0.1, ramp_time, 0.1], [0, 10, 10, 0])
        .phase.uniform.piecewise_constant(
            [0.1, ramp_time / 2, ramp_time / 2, 0.1], [0, 0, np.pi, np.pi]
        )
        .assign(ramp_time=3.0)
        .batch_assign(r=np.linspace(4, 10, 11).tolist())
        .bloqade.python()
        .run(10000, cache_matrices=True, blockade_radius=6.0)
        .report()
        .bitstrings()
    )


def test_integration_3():
    ramp_time = var("ramp_time")
    (
        start.add_position((0, 0))
        .add_position((0, 5.0))
        .scale("r")
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, ramp_time, 0.1], [-100, -100, 100, 100]
        )
        .amplitude.uniform.piecewise_linear([0.1, ramp_time, 0.1], [0, 10, 10, 0])
        .phase.uniform.piecewise_constant(
            [0.1, ramp_time / 2, ramp_time / 2, 0.1], [0, 0, np.pi, np.pi]
        )
        .amplitude.var("rabi_mask")
        .fn(lambda t: 4 * np.sin(3 * t), ramp_time + 0.2)
        .assign(ramp_time=3.0, rabi_mask=[10.0, 0.1])
        .batch_assign(r=np.linspace(4, 10, 11).tolist())
        .bloqade.python()
        .run(10000, cache_matrices=True, blockade_radius=6.0)
        .report()
        .bitstrings()
    )


def test_integration_4():
    ramp_time = var("ramp_time")
    (
        start.add_position((0, 0))
        .add_position((0, 5.0))
        .scale("r")
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, ramp_time, 0.1], [-100, -100, 100, 100]
        )
        .amplitude.uniform.piecewise_linear([0.1, ramp_time, 0.1], [0, 10, 10, 0])
        .amplitude.var("rabi_mask")
        .fn(lambda t: 4 * np.sin(3 * t), ramp_time + 0.2)
        .amplitude.location(1)
        .linear(0.0, 1.0, ramp_time + 0.2)
        .assign(ramp_time=3.0, rabi_mask=[10.0, 0.1])
        .batch_assign(r=np.linspace(4, 10, 11).tolist())
        .bloqade.python()
        .run(10000, cache_matrices=True, blockade_radius=6.0)
        .report()
        .bitstrings()
    )


def test_integration_5():
    ramp_time = var("ramp_time")
    (
        start.add_position((0, 0))
        .add_position((0, 5.0))
        .scale("r")
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, ramp_time, 0.1], [-100, -100, 100, 100]
        )
        .amplitude.uniform.piecewise_linear([0.1, ramp_time, 0.1], [0, 10, 10, 0])
        .phase.location(1)
        .linear(0.0, 1.0, ramp_time + 0.2)
        .assign(ramp_time=3.0)
        .batch_assign(r=np.linspace(4, 10, 11).tolist())
        .bloqade.python()
        .run(10000, cache_matrices=True, blockade_radius=6.0)
        .report()
        .bitstrings()
    )


def test_integration_6():
    ramp_time = var("ramp_time")
    (
        start.add_position((0, 0))
        .add_position((0, 5.0))
        .scale("r")
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, ramp_time, 0.1], [-100, -100, 100, 100]
        )
        .location(1)
        .constant(3.0, ramp_time + 0.2)
        .location(0)
        .linear(2.0, 0, ramp_time + 0.2)
        .amplitude.uniform.piecewise_linear([0.1, ramp_time, 0.1], [0, 10, 10, 0])
        .phase.location(1)
        .linear(0.0, 1.0, ramp_time + 0.2)
        .assign(ramp_time=3.0)
        .batch_assign(r=np.linspace(4, 10, 11).tolist())
        .bloqade.python()
        .run(10000, cache_matrices=True, blockade_radius=6.0)
        .report()
        .bitstrings()
    )


def test_serialization():
    ramp_time = var("ramp_time")
    batch = (
        start.add_position((0, 0))
        .add_position((0, 5.0))
        .scale("r")
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, ramp_time, 0.1], [-100, -100, 100, 100]
        )
        .amplitude.uniform.piecewise_linear([0.1, ramp_time, 0.1], [0, 10, 10, 0])
        .amplitude.var("rabi_mask")
        .piecewise_linear([0.1, ramp_time, 0.1], [0, 10, 10, 0])
        .amplitude.location(1)
        .linear(0.0, 1.0, ramp_time + 0.2)
        .assign(ramp_time=3.0, rabi_mask=[10.0, 0.1])
        .batch_assign(r=np.linspace(4, 10, 11).tolist())
        .bloqade.python()
        ._compile(100)
    )

    obj_str = dumps(batch)
    batch2 = loads(obj_str)
    assert isinstance(batch2, type(batch))



from functools import reduce
from bloqade.builder.base import Builder
from bloqade.builder.typing import ScalarType
from bloqade.builder.route import WaveformRoute

from beartype import beartype
from beartype.typing import Optional, Union, List, Callable

import bloqade.ir as ir


class WaveformAttachable(Builder):
    @beartype
    def linear(
        self, start: ScalarType, stop: ScalarType, duration: ScalarType
    ) -> "Linear":
        """
        Append/assign a linear waveform to the current location.

        Args:
            start (ScalarType Union[float, str]): The start value of the waveform
            stop (ScalarType Union[float, str]): The stop value of the waveform
            duration (ScalarType Union[float, str]): The duration of the waveform

        Examples:
            - specify a linear waveform for (spatial) uniform rydberg detuning
            from 0 to 1 in 0.5 us.

            >>> node = bloqade.start.rydberg.detuning.uniform
            >>> node = node.linear(start=0,stop=1,duration=0.5)

        Possible Next:

        - Possible Next <Location>:

            -> `.location(int)`
                :: creating new channel to address another location(s)

        - Possible Next <WaveForm:: current>:

            -> `.slice()`
                :: slice current waveform

            -> `.record(str)`
                :: record the value of waveform at current time

        - Possible Next <WaveForm:: append>:

            :: Append waveform into current channel

            -> `.linear()`

            -> `.constant()`

            -> `.ploy()`

            -> `.apply()`

            -> `.piecewise_linear()`

            -> `.piecewise_constant()`

            -> `.fn()`

        - Possible Next <LevelCoupling>:

            -> `.rydberg`
                :: Create/Switch to new rydberg level coupling channel

            -> `.hyperfine`
                :: Create/Switch to new hyperfine level coupling channel

        - Possible Next <Emit:: Linking Vars>:

            -> `.assign()`
                :: assign varialbe an actual value/number

            -> `.batch_assign()`
                :: create batch job with different
                sets of values assign to each variable.

        - Possible Next <Backend>:

            -> `.quera`
                :: specify QuEra backend

            -> `.braket`
                :: specify QuEra backend

        """
        return Linear(start, stop, duration, self)

    @beartype
    def constant(self, value: ScalarType, duration: ScalarType) -> "Constant":
        """
        Append/assign a constant waveform to the current location.

        Args:
            value (ScalarType Union[float, str]): The value of the waveform
            duration (ScalarType Union[float, str]): The duration of the waveform

        Examples:
            - specify a constant waveform of value 1 with duration 0.5
            for (spatial) uniform rydberg detuning

            >>> node = bloqade.start.rydberg.detuning.uniform
            >>> node = node.constant(value=1,duration=0.5)

        Possible Next:

        - Possible Next <Location>:

            -> `.location(int)`
                :: creating new channel to address another location(s)

        - Possible Next <WaveForm:: current>:

            -> `.slice()`
                :: slice current waveform

            -> `.record(str)`
                :: record the value of waveform at current time

        - Possible Next <WaveForm:: append>:

            :: Append waveform into current channel

            -> `.linear()`

            -> `.constant()`

            -> `.ploy()`

            -> `.apply()`

            -> `.piecewise_linear()`

            -> `.piecewise_constant()`

            -> `.fn()`

        - Possible Next <LevelCoupling>:

            -> `.rydberg`
                :: Create/Switch to new rydberg level coupling channel

            -> `.hyperfine`
                :: Create/Switch to new hyperfine level coupling channel

        - Possible Next <Emit:: Linking Vars>:

            -> `.assign()`
                :: assign varialbe an actual value/number

            -> `.batch_assign()`
                :: create batch job with different sets
                of values assign to each variable.

        - Possible Next <Backend>:

            -> `.quera`
                :: specify QuEra backend

            -> `.braket`
                :: specify QuEra backend

        """
        return Constant(value, duration, self)

    @beartype
    def poly(self, coeffs: List[ScalarType], duration: ScalarType) -> "Poly":
        """
        Append/assign a waveform with polynomial profile to the current location.
        with form:

            wv(t) = coeffs[0] + coeffs[1]*t + coeffs[2]*t^2 + ... + coeffs[n]*t^n

        Args:
            coeffs (ScalarType Union[float, str]): The coefficients of the polynomial
            duration (ScalarType Union[float, str]): The duration of the waveform

        Examples:
            - specify a second order polynomial with duration 0.5
            for (spatial) uniform rydberg detuning

            >>> node = bloqade.start.rydberg.detuning.uniform
            >>> node = node.poly(coeffs=[1,2,3],duration=0.5)

        Possible Next:

        - Possible Next <Location>:

            -> `.location(int)`
                :: creating new channel to address another location(s)

        - Possible Next <WaveForm:: current>:

            -> `.slice()`
                :: slice current waveform

            -> `.record(str)`
                :: record the value of waveform at current time

        - Possible Next <WaveForm:: append>:

            :: Append waveform into current channel

            -> `.linear()`

            -> `.constant()`

            -> `.ploy()`

            -> `.apply()`

            -> `.piecewise_linear()`

            -> `.piecewise_constant()`

            -> `.fn()`

        - Possible Next <LevelCoupling>:

            -> `.rydberg`
                :: Create/Switch to new rydberg level coupling channel

            -> `.hyperfine`
                :: Create/Switch to new hyperfine level coupling channel


        - Possible Next <Emit:: Linking Vars>:

            -> `.assign()`
                :: assign varialbe an actual value/number

            -> `.batch_assign()`
                :: create batch job with different sets
                of values assign to each variable.

        - Possible Next <Backend>:

            -> `.quera`
                :: specify QuEra backend

            -> `.braket`
                :: specify QuEra backend

        """
        return Poly(coeffs, duration, self)

    @beartype
    def apply(self, wf: ir.Waveform) -> "Apply":
        """
        Apply a pre-defined waveform to the current location.

        Args:
            wf (ir.Waveform): the waveform

        Examples:
            - apply a pre-defined waveform object to current sequence.

            >>> node = bloqade.start.rydberg.detuning.uniform
            >>> wv = bloqade.ir.Linear(0,10,0.5)
            >>> node = node.apply(wv)

        Possible Next:

        - Possible Next <Location>:

            -> `.location(int)`
                :: creating new channel to address another location(s)

        - Possible Next <WaveForm:: current>:

            -> `.slice()`
                :: slice current waveform

            -> `.record(str)`
                :: record the value of waveform at current time

        - Possible Next <WaveForm:: append>:

            :: Append waveform into current channel

            -> `.linear()`

            -> `.constant()`

            -> `.ploy()`

            -> `.apply()`

            -> `.piecewise_linear()`

            -> `.piecewise_constant()`

            -> `.fn()`

        - Possible Next <LevelCoupling>:

            -> `.rydberg`
                :: Create/Switch to new rydberg level coupling channel

            -> `.hyperfine`
                :: Create/Switch to new hyperfine level coupling channel


        - Possible Next <Emit:: Linking Vars>:

            -> `.assign()`
                :: assign varialbe an actual value/number

            -> `.batch_assign()`
                :: create batch job with different sets
                of values assign to each variable.

        - Possible Next <Backend>:

            -> `.quera`
                :: specify QuEra backend

            -> `.braket`
                :: specify QuEra backend

        """
        return Apply(wf, self)

    @beartype
    def piecewise_linear(
        self, durations: List[ScalarType], values: List[ScalarType]
    ) -> "PiecewiseLinear":
        """
        Append/assign a piecewise linear waveform to the current location.
        The durations should have number of elements = len(values) - 1.

        This function create a waveform by connecting `values[i], values[i+1]`
        with linear segments.

        Args:
            durations (List[ScalarType]): The durations of each linear segment
            values (List[ScalarType]): The values of each linear segment

        Examples:
            - specify a piecewise linear of [0,1,1,0] with duration [0.1,3.8,0.1]
            for (spatial) uniform rydberg detuning.

            >>> node = bloqade.start.rydberg.detuning.uniform
            >>> node = node.piecewise_linear(values=[0,1,1,0],durations=[0.1,3.8,0.1])

        Note:
            ScalarType can be either float or str.

        Possible Next:

        - Possible Next <Location>:

            -> `.location(int)`
                :: creating new channel to address another location(s)

        - Possible Next <WaveForm:: current>:

            -> `.slice()`
                :: slice current waveform

            -> `.record(str)`
                :: record the value of waveform at current time

        - Possible Next <WaveForm:: append>:

            :: Append waveform into current channel

            -> `.linear()`

            -> `.constant()`

            -> `.ploy()`

            -> `.apply()`

            -> `.piecewise_linear()`

            -> `.piecewise_constant()`

            -> `.fn()`

        - Possible Next <LevelCoupling>:

            -> `.rydberg`
                :: Create/Switch to new rydberg level coupling channel

            -> `.hyperfine`
                :: Create/Switch to new hyperfine level coupling channel


        - Possible Next <Emit:: Linking Vars>:

            -> `.assign()`
                :: assign varialbe an actual value/number

            -> `.batch_assign()`
                :: create batch job with different sets
                of values assign to each variable.

        - Possible Next <Backend>:

            -> `.quera`
                :: specify QuEra backend

            -> `.braket`
                :: specify QuEra backend

        """
        return PiecewiseLinear(durations, values, self)

    @beartype
    def piecewise_constant(
        self, durations: List[ScalarType], values: List[ScalarType]
    ) -> "PiecewiseConstant":
        """
        Append/assign a piecewise constant waveform to the current location.
        The durations should have number of elements = len(values).

        This function create a waveform of piecewise_constant of
        `values[i]` with duration `durations[i]`.

        Args:
            durations (List[ScalarType]): The durations of each constant segment
            values (List[ScalarType]): The values of each constant segment

        Note:
            ScalarType can be either float or str.

        Examples:
            - specify a piecewise constant of [0.5,1.5] with duration [0.1,3.8]
            for (spatial) uniform rydberg detuning.

            >>> node = bloqade.start.rydberg.detuning.uniform
            >>> node = node.piecewise_constant(values=[0.5,1.5],durations=[0.1,3.8])

        Possible Next:

        - Possible Next <Location>:

            -> `.location(int)`
                :: creating new channel to address another location(s)

        - Possible Next <WaveForm:: current>:

            -> `.slice()`
                :: slice current waveform

            -> `.record(str)`
                :: record the value of waveform at current time

        - Possible Next <WaveForm:: append>:

            :: Append waveform into current channel

            -> `.linear()`

            -> `.constant()`

            -> `.ploy()`

            -> `.apply()`

            -> `.piecewise_linear()`

            -> `.piecewise_constant()`

            -> `.fn()`

        - Possible Next <LevelCoupling>:

            -> `.rydberg`
                :: Create/Switch to new rydberg level coupling channel

            -> `.hyperfine`
                :: Create/Switch to new hyperfine level coupling channel


        - Possible Next <Emit:: Linking Vars>:

            -> `.assign()`
                :: assign varialbe an actual value/number

            -> `.batch_assign()`
                :: create batch job with different sets
                of values assign to each variable.


        - Possible Next <Backend>:

            -> `.quera`
                :: specify QuEra backend

            -> `.braket`
                :: specify QuEra backend

        """
        return PiecewiseConstant(durations, values, self)

    @beartype
    def fn(self, fn: Callable, duration: ScalarType) -> "Fn":
        """
        Append/assign a waveform defined by a python function to the current location.

        This function create a waveform with user-defined
        python function `fn(t)` with duration `duration`.

        Args:
            fn (Callable): The python function defining the waveform
            duration (ScalarType): The durations of each constant segment

        Note:
            - ScalarType can be either float or str.
            - The python function should take a single argument `t` and return a float.


        Examples:
            - create a cosine waveform with duration 0.5
            for (spatial) uniform rydberg detuning.

            >>> import numpy as np
            >>> def my_fn(t):
            >>>     return np.cos(2*np.pi*t)
            >>> node = bloqade.start.rydberg.detuning.uniform
            >>> node = node.fn(my_fn,duration=0.5)

        Possible Next:

        - Possible Next <Location>:

            -> `.location(int)`
                :: creating new channel to address another location(s)

        - Possible Next <WaveForm:: current>:

            -> `.slice()`
                :: slice current waveform

            -> `.record(str)`
                :: record the value of waveform at current time

            -> `.sample()`
                :: sample current callable at given time points

        - Possible Next <WaveForm:: append>:

            :: Append waveform into current channel

            -> `.linear()`

            -> `.constant()`

            -> `.ploy()`

            -> `.apply()`

            -> `.piecewise_linear()`

            -> `.piecewise_constant()`

            -> `.fn()`

        - Possible Next <LevelCoupling>:

            -> `.rydberg`
                :: Create/Switch to new rydberg level coupling channel

            -> `.hyperfine`
                :: Create/Switch to new hyperfine level coupling channel


        - Possible Next <Emit:: Linking Vars>:

            -> `.assign()`
                :: assign varialbe an actual value/number

            -> `.batch_assign()`
                :: create batch job with different sets
                of values assign to each variable.


        - Possible Next <Backend>:

            -> `.quera`
                :: specify QuEra backend

            -> `.braket`
                :: specify QuEra backend

        """
        return Fn(fn, duration, self)


# NOTE: waveform can refer previous pulse notes
#       or continue to specify pragma options
class Waveform(WaveformRoute, WaveformAttachable):
    pass


# mixin for slice and record
class Slicible:
    @beartype
    def slice(
        self,
        start: Optional[ScalarType] = None,
        stop: Optional[ScalarType] = None,
    ) -> "Slice":
        """
        Slice current waveform

        Possible Next:

        - Possible Next <Location>:

            -> `.location(int)`
                :: creating new channel to address
                another location(s)

        - Possible Next <WaveForm:: current>:

            -> `.record(str)`
                :: record the value of waveform at current time

        - Possible Next <WaveForm:: append>:

            :: Append waveform into current channel

            -> `.linear()`

            -> `.constant()`

            -> `.ploy()`

            -> `.apply()`

            -> `.piecewise_linear()`

            -> `.piecewise_constant()`

            -> `.fn()`

        - Possible Next <LevelCoupling>:

            -> `.rydberg`
                :: Create/Switch to new rydberg level coupling channel

            -> `.hyperfine`
                :: Create/Switch to new hyperfine level coupling channel


        - Possible Next <Emit:: Linking Vars>:

            -> `.assign()`
                :: assign varialbe an actual value/number

            -> `.batch_assign()`
                :: create batch job with different sets
                of values assign to each variable.


        - Possible Next <Backend>:

            -> `.quera`
                :: specify QuEra backend

            -> `.braket`
                :: specify QuEra backend
        """
        return Slice(start, stop, self)


class Recordable:
    @beartype
    def record(self, name: str) -> "Record":
        """
        Record the value of the current waveform to a variable.

        Possible Next:

        - Possible Next <Location>:

            -> `.location(int)`
                :: creating new channel to address
                another location(s)

        - Possible Next <WaveForm:: current>:

            -> `.slice()`
                :: slice the current waveform

            -> `.record(str)`
                :: record the value of waveform at current time

        - Possible Next <WaveForm:: append>:

            :: Append waveform into current channel

            -> `.linear()`

            -> `.constant()`

            -> `.ploy()`

            -> `.apply()`

            -> `.piecewise_linear()`

            -> `.piecewise_constant()`

            -> `.fn()`

        - Possible Next <LevelCoupling>:

            -> `.rydberg`
                :: Create/Switch to new rydberg level coupling channel

            -> `.hyperfine`
                :: Create/Switch to new hyperfine level coupling channel

        - Possible Next <Emit:: Linking Vars>:

            -> `.assign()`
                :: assign varialbe an actual value/number

            -> `.batch_assign()`
                :: create batch job with different sets
                of values assign to each variable.

        - Possible Next <Backend>:

            -> `.quera`
                :: specify QuEra backend

            -> `.braket`
                :: specify QuEra backend

        """
        return Record(name, self)


class WaveformPrimitive(Waveform, Slicible, Recordable):
    def __bloqade_ir__(self):
        raise NotImplementedError


class Linear(WaveformPrimitive):
    __match_args__ = ("_start", "_stop", "_duration", "__parent__")

    def __init__(
        self,
        start: ScalarType,
        stop: ScalarType,
        duration: ScalarType,
        parent: Optional[Builder] = None,
    ) -> None:
        super().__init__(parent)
        self._start = ir.cast(start)
        self._stop = ir.cast(stop)
        self._duration = ir.cast(duration)

    def __bloqade_ir__(self) -> ir.Linear:
        return ir.Linear(start=self._start, stop=self._stop, duration=self._duration)


class Constant(WaveformPrimitive):
    __match_args__ = ("_value", "_duration", "__parent__")

    def __init__(
        self, value: ScalarType, duration: ScalarType, parent: Optional[Builder] = None
    ) -> None:
        super().__init__(parent)
        self._value = ir.cast(value)
        self._duration = ir.cast(duration)

    def __bloqade_ir__(self) -> ir.Constant:
        return ir.Constant(value=self._value, duration=self._duration)


class Poly(WaveformPrimitive):
    __match_args__ = ("_coeffs", "_duration", "__parent__")

    def __init__(
        self,
        coeffs: List[ScalarType],
        duration: ScalarType,
        parent: Optional[Builder] = None,
    ) -> None:
        super().__init__(parent)
        self._coeffs = ir.cast(coeffs)
        self._duration = ir.cast(duration)

    def __bloqade_ir__(self):
        return ir.Poly(coeffs=self._coeffs, duration=self._duration)


class Apply(WaveformPrimitive):
    __match_args__ = ("_wf", "__parent__")

    def __init__(self, wf: ir.Waveform, parent: Optional[Builder] = None):
        super().__init__(parent)
        self._wf = wf

    def __bloqade_ir__(self):
        return self._wf


class PiecewiseLinear(WaveformPrimitive):
    __match_args__ = ("_durations", "_values", "__parent__")

    def __init__(
        self,
        durations: List[ScalarType],
        values: List[ScalarType],
        parent: Optional[Builder] = None,
    ):
        assert (
            len(durations) == len(values) - 1
        ), "durations and values must be the same length"

        super().__init__(parent)
        self._durations = list(map(ir.cast, durations))
        self._values = list(map(ir.cast, values))

    def __bloqade_ir__(self):
        iter = zip(self._values[:-1], self._values[1:], self._durations)
        wfs = [ir.Linear(start=v0, stop=v1, duration=t) for v0, v1, t in iter]

        return reduce(lambda a, b: a.append(b), wfs)


class PiecewiseConstant(WaveformPrimitive):
    __match_args__ = ("_durations", "_values", "__parent__")

    def __init__(
        self,
        durations: List[ScalarType],
        values: List[ScalarType],
        parent: Optional[Builder] = None,
    ) -> None:
        assert len(durations) == len(
            values
        ), "durations and values must be the same length"
        super().__init__(parent)
        self._durations = list(map(ir.cast, durations))
        self._values = list(map(ir.cast, values))

    def __bloqade_ir__(self):
        iter = zip(self._values, self._durations)
        wfs = [ir.Constant(value=v, duration=t) for v, t in iter]
        return reduce(lambda a, b: a.append(b), wfs)


class Fn(WaveformPrimitive):
    __match_args__ = ("_fn", "_duration", "__parent__")

    def __init__(
        self,
        fn: Callable,
        duration: ScalarType,
        parent: Optional[Builder] = None,
    ) -> None:
        super().__init__(parent)
        self._fn = fn
        self._duration = ir.cast(duration)

    @beartype
    def sample(
        self, dt: ScalarType, interpolation: Union[ir.Interpolation, str, None] = None
    ):
        return Sample(dt, interpolation, self)

    def __bloqade_ir__(self):
        return ir.PythonFn(self._fn, self._duration)


# NOTE: no double-slice or double-record
class Slice(Waveform, Recordable):
    __match_args__ = ("_start", "_stop", "__parent__")

    def __init__(
        self,
        start: Optional[ScalarType] = None,
        stop: Optional[ScalarType] = None,
        parent: Optional[Builder] = None,
    ) -> None:
        super().__init__(parent)
        # NOTE: this should no raise for None
        self._start = ir.scalar.trycast(start)
        self._stop = ir.scalar.trycast(stop)


class Record(Waveform, Slicible):  # record should not be sliceable
    __match_args__ = ("_name", "__parent__")

    def __init__(
        self,
        name: str,
        parent: Optional[Builder] = None,
    ) -> None:
        super().__init__(parent)
        self._name = ir.var(name)


class Sample(Slicible, Recordable, WaveformRoute):
    __match_args__ = ("_dt", "_interpolation", "__parent__")

    def __init__(
        self,
        dt: ScalarType,
        interpolation: Union[ir.Interpolation, str, None],
        parent: Builder,
    ) -> None:
        super().__init__(parent)
        self._dt = ir.cast(dt)
        if interpolation is None:
            self._interpolation = None
        else:
            self._interpolation = ir.Interpolation(interpolation)



from dataclasses import dataclass
from numpy.typing import NDArray
from typing import TYPE_CHECKING
import numpy as np
from enum import Enum

if TYPE_CHECKING:
    from .emulator import Register
    from .atom_type import AtomType


class SpaceType(str, Enum):
    FullSpace = "full_space"
    SubSpace = "sub_space"


@dataclass(frozen=True)
class Space:
    space_type: SpaceType
    atom_type: "AtomType"
    geometry: "Register"
    configurations: NDArray

    @classmethod
    def create(cls, register: "Register"):
        sites = register.sites
        n_atom = len(sites)
        atom_type = register.atom_type
        blockade_radius = register.blockade_radius
        Ns = atom_type.n_level**n_atom

        check_atoms = []

        for index_1, site_1 in enumerate(sites):
            site_1 = np.asarray(site_1)
            atoms = []
            for index_2, site_2 in enumerate(sites[index_1 + 1 :], index_1 + 1):
                site_2 = np.asarray(site_2)
                if np.linalg.norm(site_1 - site_2) <= blockade_radius:
                    atoms.append(index_2)

            check_atoms.append(atoms)

        configurations = np.arange(Ns, dtype=np.min_scalar_type(Ns - 1))

        if all(len(sub_list) == 0 for sub_list in check_atoms):
            return Space(SpaceType.FullSpace, atom_type, sites, configurations)

        for index_1, indices in enumerate(check_atoms):
            # get which configurations are in rydberg state for the current index.
            rydberg_configs_1 = atom_type.is_rydberg_at(configurations, index_1)
            for index_2 in indices:  # loop over neighbors within blockade radius
                # get which configus have the neighbor with a rydberg excitation
                rydberg_configs_2 = atom_type.is_rydberg_at(configurations, index_2)
                # get which states do not violate constraint
                mask = np.logical_not(
                    np.logical_and(rydberg_configs_1, rydberg_configs_2)
                )
                # remove states that violate blockade constraint
                configurations = configurations[mask]
                rydberg_configs_1 = rydberg_configs_1[mask]

        min_int_type = np.min_scalar_type(configurations[-1])
        # defauly to 32 bit if smaller than 32 bit
        config_type = np.result_type(min_int_type, np.uint32)
        configurations = configurations.astype(config_type)

        return cls(SpaceType.SubSpace, atom_type, sites, configurations)

    @property
    def index_type(self) -> np.dtype:
        if self.size < np.iinfo(np.int32).max:
            return np.int32
        else:
            return np.int64

    @property
    def size(self) -> int:
        return self.configurations.size

    @property
    def n_atoms(self) -> int:
        return len(self.geometry)

    @property
    def state_type(self) -> np.dtype:
        return self.configurations.dtype

    def is_rydberg_at(self, index: int) -> NDArray:
        return self.atom_type.is_rydberg_at(self.configurations, index)

    def is_state_at(self, index: int, state: int):
        return self.atom_type.is_state_at(self.configurations, index, state)

    def swap_state_at(self, index: int, state_1: int, state_2: int) -> NDArray:
        row_indices, col_config = self.atom_type.swap_state_at(
            self.configurations, index, state_1, state_2
        )
        if self.space_type == SpaceType.FullSpace:
            return (row_indices, col_config)
        else:
            col_indices = np.searchsorted(self.configurations, col_config)

            mask = col_indices < self.size
            mask[mask] = col_config[mask] == self.configurations[col_indices[mask]]

            if not np.all(mask):
                if isinstance(row_indices, slice):
                    return mask, col_indices[mask]
                else:
                    return row_indices[mask], col_indices[mask]
            else:
                return row_indices, col_indices

    def transition_state_at(self, index: int, fro: int, to: int) -> NDArray:
        row_indices, col_config = self.atom_type.transition_state_at(
            self.configurations, index, fro, to
        )
        if self.space_type == SpaceType.FullSpace:
            return (row_indices, col_config)
        else:
            col_indices = np.searchsorted(self.configurations, col_config)
            row_indices = np.argwhere(row_indices).ravel()

            mask = col_indices < self.size
            col_indices = col_indices[mask]
            row_indices = row_indices[mask]

            mask = col_config == self.configurations[col_indices]
            col_indices = col_indices[mask]
            row_indices = row_indices[mask]

            return (row_indices, col_indices)

    def fock_state_to_index(self, fock_state: str) -> int:
        state_int = self.atom_type.string_to_integer(fock_state)
        if self.space_type is SpaceType.FullSpace:
            return state_int
        else:
            index = np.searchsorted(self.configurations, state_int)

            if state_int != self.configurations[index]:
                raise ValueError(
                    "state: {fock_state} not in rydberg blockade subspace."
                )

            return index

    def index_to_fock_state(self, index: int) -> str:
        if self.space_type is SpaceType.FullSpace:
            return self.atom_type.integer_to_string(index, self.n_atoms)
        else:
            return self.atom_type.integer_to_string(
                self.configurations[index], self.n_atoms
            )

    def zero_state(self, dtype=np.float64) -> NDArray:
        state = np.zeros(self.size, dtype=dtype)
        state[0] = 1.0
        return state

    def sample_state_vector(
        self, state_vector: NDArray, n_samples: int, project_hyperfine: bool = True
    ) -> NDArray:
        from .atom_type import ThreeLevelAtomType

        p = np.abs(state_vector) ** 2
        sampled_configs = np.random.choice(self.configurations, size=n_samples, p=p)

        sample_fock_states = np.empty((n_samples, self.n_atoms), dtype=np.uint8)

        for i in range(self.n_atoms):
            sample_fock_states[:, i] = sampled_configs % self.atom_type.n_level
            sampled_configs //= self.atom_type.n_level

        if project_hyperfine and isinstance(self.atom_type, ThreeLevelAtomType):
            sample_fock_states[sample_fock_states == 1] = 0
            sample_fock_states[sample_fock_states == 2] = 1

        return sample_fock_states

    def __str__(self):
        # TODO: update this to use unicode
        output = ""

        n_digits = len(str(self.size - 1))
        fmt = "{{index: >{}d}}. {{fock_state:s}}\n".format(n_digits)
        if self.size < 50:
            for index, state_int in enumerate(self.configurations):
                fock_state = self.atom_type.integer_to_string(state_int, self.n_atoms)
                output = output + fmt.format(index=index, fock_state=fock_state)

        else:
            for index, state_int in enumerate(self.configurations[:25]):
                fock_state = self.atom_type.integer_to_string(state_int, self.n_atoms)
                output = output + fmt.format(index=index, fock_state=fock_state)

            output += (n_digits * "  ") + "...\n"

            for index, state_int in enumerate(
                self.configurations[-25:], start=self.size - 25
            ):
                fock_state = self.atom_type.integer_to_string(state_int, self.n_atoms)
                output = output + fmt.format(index=index, fock_state=fock_state)

        return output



from beartype.typing import Optional
from bloqade.builder.typing import ScalarType
from bloqade.builder.base import Builder
from bloqade.builder.backend import BackendRoute
from bloqade.builder.parse.trait import Parse
from bloqade.ir import cast


class Parallelize(BackendRoute, Parse):
    __match_args__ = ("_cluster_spacing", "__parent__")

    def __init__(
        self, cluster_spacing: ScalarType, parent: Optional[Builder] = None
    ) -> None:
        super().__init__(parent)
        self._cluster_spacing = cast(cluster_spacing)



from typing import Any, Dict, TextIO, Type, Union
from bloqade.builder.base import Builder
from bloqade.codegen.common.json import BloqadeIRSerializer, BloqadeIRDeserializer
from bloqade.builder.start import ProgramStart
from bloqade.builder.sequence_builder import SequenceBuilder

# from bloqade.builder.base import Builder
from bloqade.builder.backend.bloqade import BloqadeDeviceRoute, BloqadeService
from bloqade.builder.backend.braket import BraketDeviceRoute, BraketService
from bloqade.builder.backend.quera import QuEraDeviceRoute, QuEraService
from bloqade.builder.spatial import Location, Scale, Var, Uniform
from bloqade.builder.waveform import (
    Linear,
    Constant,
    Poly,
    Apply,
    Slice,
    Record,
    Sample,
    PiecewiseLinear,
    PiecewiseConstant,
)
from bloqade.builder.field import Detuning, Rabi, RabiAmplitude, RabiPhase
from bloqade.builder.coupling import Rydberg, Hyperfine
from bloqade.builder.parallelize import Parallelize
from bloqade.builder.assign import Assign, BatchAssign
from bloqade.builder.flatten import Flatten


class BuilderSerializer(BloqadeIRSerializer):
    types = set(
        [
            ProgramStart,
            SequenceBuilder,
            BloqadeDeviceRoute,
            BloqadeService,
            BraketDeviceRoute,
            BraketService,
            QuEraDeviceRoute,
            QuEraService,
            Location,
            Scale,
            Var,
            Uniform,
            Linear,
            Constant,
            Poly,
            Apply,
            Slice,
            Record,
            Sample,
            PiecewiseLinear,
            PiecewiseConstant,
            Detuning,
            Rabi,
            RabiAmplitude,
            RabiPhase,
            Rydberg,
            Hyperfine,
            Parallelize,
            Assign,
            BatchAssign,
            Flatten,
        ]
    )

    def default(self, obj):
        if isinstance(obj, ProgramStart) and type(obj) is ProgramStart:
            return "program_start"

        if type(obj) in self.types:
            return {
                type(obj).__name__: {
                    arg: getattr(obj, arg) for arg in obj.__match_args__
                }
            }


class BuilderDeserializer(BloqadeIRDeserializer):
    methods: Dict[str, Type] = {
        "ProgramStart": ProgramStart,
        "SequenceBuilder": SequenceBuilder,
        "BloqadeDeviceRoute": BloqadeDeviceRoute,
        "BloqadeService": BloqadeService,
        "BraketDeviceRoute": BraketDeviceRoute,
        "BraketService": BraketService,
        "QuEraDeviceRoute": QuEraDeviceRoute,
        "QuEraService": QuEraService,
        "Location": Location,
        "Scale": Scale,
        "Var": Var,
        "Uniform": Uniform,
        "Linear": Linear,
        "Constant": Constant,
        "Poly": Poly,
        "Apply": Apply,
        "Slice": Slice,
        "Record": Record,
        "Sample": Sample,
        "PiecewiseLinear": PiecewiseLinear,
        "PiecewiseConstant": PiecewiseConstant,
        "Detuning": Detuning,
        "Rabi": Rabi,
        "RabiAmplitude": RabiAmplitude,
        "RabiPhase": RabiPhase,
        "Rydberg": Rydberg,
        "Hyperfine": Hyperfine,
        "Parallelize": Parallelize,
        "Assign": Assign,
        "BatchAssign": BatchAssign,
        "Flatten": Flatten,
    }

    def object_hook(self, obj: Dict[str, Any]):
        if isinstance(obj, dict) and len(obj) == 1:
            ((head, options),) = obj.items()
            if head in self.methods:
                return self.methods[head](**options)
        elif obj == "program_start":
            return ProgramStart(None)

        return super().object_hook(obj)


def load_program(filename_or_io: Union[str, TextIO]) -> Builder:
    import json

    if isinstance(filename_or_io, str):
        with open(filename_or_io, "r") as f:
            return json.load(f, object_hook=BuilderDeserializer().object_hook)
    else:
        return json.load(filename_or_io, object_hook=BuilderDeserializer().object_hook)


def save_program(filename_or_io: Union[str, TextIO], program: Builder) -> None:
    import json

    if isinstance(filename_or_io, str):
        with open(filename_or_io, "w") as f:
            json.dump(program, f, cls=BuilderSerializer)
    else:
        json.dump(program, filename_or_io, cls=BuilderSerializer)



from bloqade.ir.control.waveform import Waveform, Linear, Constant
from bloqade.builder.typing import ScalarType
from beartype import beartype
from beartype.typing import List

# this part only for manually build sequences.


@beartype
def linear(duration: ScalarType, start: ScalarType, stop: ScalarType):
    return Linear(start, stop, duration)


@beartype
def constant(duration: ScalarType, value: ScalarType):
    return Constant(value, duration)


@beartype
def piecewise_linear(durations: List[ScalarType], values: List[ScalarType]) -> Waveform:
    pwl_wf = None
    for duration, start, stop in zip(durations, values[:-1], values[1:]):
        if pwl_wf is None:
            pwl_wf = Linear(start, stop, duration)
        else:
            pwl_wf = pwl_wf.append(Linear(start, stop, duration))

    return pwl_wf


@beartype
def piecewise_constant(
    durations: List[ScalarType], values: List[ScalarType]
) -> Waveform:
    pwc_wf = None
    for duration, value in zip(durations, values):
        if pwc_wf is None:
            pwc_wf = Constant(value, duration)
        else:
            pwc_wf = pwc_wf.append(Constant(value, duration))

    return pwc_wf



from bloqade.emulate.ir.atom_type import TwoLevelAtom
from bloqade.emulate.ir.emulator import (
    EmulatorProgram,
    Register,
    RabiOperatorData,
    RabiTerm,
    Fields,
    DetuningTerm,
    DetuningOperatorData,
    RabiOperatorType,
)
from bloqade.emulate.codegen.emulator_ir import (
    EmulatorProgramCodeGen,
    CompiledWaveform,
    LevelCoupling,
)
from bloqade import start
from bloqade.ir.control.sequence import rydberg
from bloqade.ir.control.pulse import detuning, rabi
from bloqade.ir.control.field import Uniform
from decimal import Decimal


def test_codegen_detuning():
    program = (
        start.add_position((0, 0))
        .add_position((0, 5))
        .rydberg.detuning.uniform.piecewise_linear([0.1, 0.8, 0.1], [-10, -10, 10, 10])
        .parse_circuit()
    )

    detuning_waveform = program.sequence.pulses[rydberg].fields[detuning].value[Uniform]

    compiled_waveform = CompiledWaveform({}, detuning_waveform)

    geometry = Register(
        TwoLevelAtom,
        [(Decimal("0"), Decimal("0")), (Decimal("0"), Decimal("5"))],
        Decimal("0"),
    )

    detuning_term = DetuningTerm(
        DetuningOperatorData({i: Decimal("1") for i in range(len(geometry))}),
        compiled_waveform,
    )

    rydberg_drive = Fields(detuning=[detuning_term], rabi=[])

    drives = {LevelCoupling.Rydberg: rydberg_drive}

    expected_emulator_ir = EmulatorProgram(
        register=geometry, duration=1.0, pulses=drives
    )

    emulator_ir = EmulatorProgramCodeGen().emit(program)

    print(emulator_ir)
    print(expected_emulator_ir)

    assert emulator_ir == expected_emulator_ir


def test_codegen_detuning_and_rabi():
    program = (
        start.add_position((0, 0))
        .add_position((0, 5))
        .rydberg.detuning.uniform.piecewise_linear([0.1, 0.8, 0.1], [-10, -10, 10, 10])
        .amplitude.uniform.piecewise_linear([0.1, 0.8, 0.1], [0, 10, 10, 0])
        .parse_circuit()
    )

    detuning_waveform = program.sequence.pulses[rydberg].fields[detuning].value[Uniform]
    amplitude_waveform = (
        program.sequence.pulses[rydberg].fields[rabi.amplitude].value[Uniform]
    )

    compiled_detuning = CompiledWaveform({}, detuning_waveform)
    compiled_amplitude = CompiledWaveform({}, amplitude_waveform)

    geometry = Register(
        TwoLevelAtom,
        [(Decimal("0"), Decimal("0")), (Decimal("0"), Decimal("5"))],
        Decimal("0"),
    )

    detuning_term = DetuningTerm(
        DetuningOperatorData({i: Decimal("1") for i in range(len(geometry))}),
        compiled_detuning,
    )

    rabi_term = RabiTerm(
        RabiOperatorData(
            RabiOperatorType.RabiSymmetric,
            {i: Decimal("1") for i in range(len(geometry))},
        ),
        compiled_amplitude,
        None,
    )

    rydberg_drive = Fields(detuning=[detuning_term], rabi=[rabi_term])

    drives = {LevelCoupling.Rydberg: rydberg_drive}

    expected_emulator_ir = EmulatorProgram(
        register=geometry, duration=1.0, pulses=drives
    )

    emulator_ir = EmulatorProgramCodeGen().emit(program)

    print(emulator_ir)
    print(expected_emulator_ir)

    assert emulator_ir == expected_emulator_ir



# from numbers import Real
from typing import TYPE_CHECKING, Union
from bloqade.visualization import display_ir

if TYPE_CHECKING:
    from bloqade.ir.location.base import AtomArrangement, ParallelRegister
    from bloqade.ir import Sequence


# NOTE: this is just a dummy type bundle geometry and sequence
#       information together and forward them to backends.
class AnalogCircuit:
    """AnalogCircuit is a dummy type that bundle register and sequence together."""

    def __init__(
        self,
        register: Union["AtomArrangement", "ParallelRegister"],
        sequence: "Sequence",
    ):
        self._sequence = sequence
        self._register = register

    @property
    def register(self):
        """Get the register of the program.

        Returns:
            register (Union["AtomArrangement", "ParallelRegister"])

        Note:
            If the program is built with
            [`parallelize()`][bloqade.builder.emit.Emit.parallelize],
            The the register will be a
            [`ParallelRegister`][bloqade.ir.location.base.ParallelRegister].
            Otherwise it will be a
            [`AtomArrangement`][bloqade.ir.location.base.AtomArrangement].
        """
        return self._register

    @property
    def sequence(self):
        """Get the sequence of the program.


        Returns:
            Sequence: the sequence of the program.
                See also [`Sequence`][bloqade.ir.control.sequence.Sequence].

        """
        return self._sequence

    def __eq__(self, other):
        if isinstance(other, AnalogCircuit):
            return (self.register == other.register) and (
                self.sequence == other.sequence
            )

        return False

    def __repr__(self):
        # TODO: add repr for static_params, batch_params and order
        out = ""
        if self._register is not None:
            out += self._register.__repr__()

        out += "\n"

        if self._sequence is not None:
            out += self._sequence.__repr__()

        return out

    def figure(self, **assignments):
        fig_reg = self._register.figure(**assignments)
        fig_seq = self._sequence.figure(**assignments)
        return fig_seq, fig_reg

    def show(self, **assignments):
        """Interactive visualization of the program

        Args:
            **assignments: assigning the instance value (literal) to the
                existing variables in the program

        """
        display_ir(self, assignments)
